#include <math.h>
#include <stdio.h>
#include <stdint.h>


/* make a large block of (4 byte alligned) memory.
 * Put the original points at the END of it.
 * Pass in a pointer to the block, the number of
 * points in the original, and the number of bytes
 * in the block. This function will recurse until
 * the memory block is full, then create the image.
 * At some point I will fix this algorithm to only
 * recurse those segments which are longer than
 * 'accuracy'.
 */

#include "koch.h"
#include "backend.h"

void write_pnm(float* pixbuf, int width, int height);
void adjust_brightness(float* pixbuf, int width, int height);

#define PUSH(stack, x, y, r, theta) {*(stack)++ = (x); *(stack)++ = (y); *(stack)++ = (r); *(stack)++ = (theta);}
#define POP(stack, x, y, r, theta) {(theta) = *(--(stack)); (r) = *(--(stack)); (y) = *(--(stack)); (x) = *(--(stack));}

#ifndef __KOCH_GUI_C_

#define BUFSIZE 200002
#define WIDTH 1152
#define HEIGHT 864
#define MAXPIXVALUE 255

void usage(char *argv[])
{
	fprintf(stderr, "usage: %s x y x y x y [x y [x y [...]]]\n", argv[0]);
	exit(1);
}

int main(int argc, char *argv[])
{
	int c, numPoints, doneSize;
	float original[MAXPOINTS*2];
	static float* buf = 0;
	static float* pixbuf = 0;
	render_args rargs;

	if(buf == 0)
	{
		buf = (float*)malloc(8 + BUFSIZE * sizeof(float));
		if(!buf)
		{
		        fprintf(stderr, "Could not allocate %d bytes\n", BUFSIZE * sizeof(float));
		        exit(1);
		}
	}

	if(pixbuf == 0)
	{
		pixbuf = (float*)calloc(WIDTH * HEIGHT, sizeof(float));
		if(!pixbuf)
		{
			fprintf(stderr, "Could not allocate %d bytes\n", WIDTH * HEIGHT);
			exit(1);
		}
	}


	if(argc < 5 || argc > MAXPOINTS*2 + 1 || !(argc % 2))
	usage(argv);

	for(c = 1; c < argc; ++c)
	{
		original[c-1] = buf[c-1] = (float) atoi(argv[c]);
		++c;
		original[c-1] = buf[c-1] = (float) atoi(argv[c]);
	}
	numPoints = c/2 - 1;



	rargs.g_width = WIDTH;
	rargs.g_height = HEIGHT;
	rargs.g_max_length = RESOLUTION;
	rargs.g_pixbuf = pixbuf;
	rargs.g_original_size = numPoints;
	rargs.tos = rargs.buf = buf;
	rargs.g_original = original;
	rargs.count = 0;

	if(rargs_setup(&rargs))
	{
		fprintf(stderr, "Can't render because one of the segments is too long\n");
		exit(14);
	}

	/* RENDER */
	while(render_3(&rargs, 200) == 0)
		;

	fprintf(stderr, "Koch has %d points.\n", rargs.count);

	/* figure out how bright it is and give it good contrast in the range 0-255 */
	adjust_brightness(pixbuf, WIDTH, HEIGHT);

	/* write to file: */
	write_pnm(pixbuf, WIDTH, HEIGHT);

	/* cleanup: */
	free(buf);
	free(pixbuf);

	return 0;
}
#endif /* !_KOCH_GUI_ */

int rargs_setup(render_args *rargs)
{
	float x, y, r, theta;
	int c;

	memcpy(rargs->buf, rargs->g_original, (rargs->g_original_size * 2 + 2) * sizeof(float));
	
	/* convert points to polar with length from first point to last == 1 */
	rec_to_norm_polar(rargs->buf, rargs->g_original_size + 1);

	for(c = 0; c < rargs->g_original_size; ++c)
	{
		/* we can't render if one of the sections is longer that the original */
		/* FIXME: make this check more sophisticated, so we can do long spirals and/or save memory */
		if(rargs->buf[c*2] > .995)
			return 1;
		rargs->g_lengths[c] = rargs->buf[c*2];
		rargs->g_angles[c] = rargs->buf[c*2+1];
	}
	
	for(c = 0; c < rargs->g_height * rargs->g_width; ++c)
		rargs->g_pixbuf[c] = 0.0;

	/* put the x and y coordinates for the last point on the stack */
	PUSH(rargs->tos, rargs->g_original[rargs->g_original_size * 2], rargs->g_original[rargs->g_original_size * 2 + 1], 0, 0);

	/* put the vector of the first point to the last on the stack */
	/* for a seccond here x and y are the width and height of the original */
	x = rargs->g_original[rargs->g_original_size * 2] - rargs->g_original[0];
	y = rargs->g_original[rargs->g_original_size * 2 + 1] - rargs->g_original[1];
	theta = patan(x, y);
	r = sqrt(x*x + y*y);
	x = rargs->g_original[0];
	y = rargs->g_original[1];
	PUSH(rargs->tos, x, y, r, theta);

	return 0;
}

/* needs a bunch of globals (original, original_size, lengths ...) */
int render_3(render_args *rargs, int enough)
{
	float x, y, r, theta;
	float x2, y2, r2, theta2;
	float x3, y3;
	int c;
	int count = 0;
	float *pixAddr;

	POP(rargs->tos, x, y, r, theta);

	while(--enough)
	{

		if(r < rargs->g_max_length)
		{
			if(x > 0 && y > 0 && x < rargs->g_width && y < rargs->g_height)
			{
				pixAddr = rargs->g_pixbuf + (int)x + ((int)y * rargs->g_width);
				*pixAddr += r;
				++(rargs->count);
			}
			if(rargs->tos == rargs->buf)
				return 1;
			POP(rargs->tos, x, y, r, theta);
		}
		else
		{
			for(c = rargs->g_original_size - 1; c > 0; --c)
			{
				theta2 = theta + rargs->g_angles[c];
				r2 = r * rargs->g_lengths[c];
				x2 = rargs->tos[-4] - cos(theta2) * r2;
				y2 = rargs->tos[-3] - sin(theta2) * r2;
				PUSH(rargs->tos, x2, y2, r2, theta2);
			}

			theta = theta + rargs->g_angles[0];
			r = r * rargs->g_lengths[0];
			x = x + cos(theta2) * r;
			y = y + sin(theta2) * r;
		}
	}

	PUSH(rargs->tos, x, y, r, theta);
	return 0;
}
	
/* look at all the pixel values and adjust the brightness (curve?) 
 * so as to get good contrast with values ranging from 0 to 255 */
void adjust_brightness(float* pixbuf, int width, int height)
{
	int c;
	int stat[256];
	float *bufend, *bufstart = pixbuf;
	int whitecount = 0;
	float factor;

	for(c = 0; c < 256; ++c)
	stat[c] = 0;

#define STAT_POW 1.1
#define STAT(a) (pow(STAT_POW, a))

	for(bufend = pixbuf + (width * height); pixbuf < bufend; ++pixbuf)
	{
		if(*pixbuf == 0)
			stat[0] += 1;
		else
			stat[(uint8_t)(log(*pixbuf + 1)/log(STAT_POW))] += 1;
	}


	for(c = 255; c > 0; c--)
	{
		whitecount += stat[c];

		//factor = log(pow(STAT(c), 1.0/254.0));

		if(whitecount > (width*height - stat[0])/80)
		{
			fprintf(stderr, "The significant bright pixels were in stat[%d]\n", c);

			for(pixbuf = bufstart, bufend = pixbuf + (width * height); pixbuf < bufend; ++pixbuf)
			{
				if(*pixbuf)
				{
					//*pixbuf = log(*pixbuf + 1) / factor;
					*pixbuf *= (255.0/STAT(c+1));
					if(*pixbuf > 254)
						*pixbuf = 254;
				}
			}
		return;
		}
	}
	
	fprintf(stderr, "ERROR: image is all black!\n");
	//exit(1);
}

void write_pnm(float* pixbuf, int width, int height)
{
	float *pixend = pixbuf + width * height;
	printf("P5\n#Created with koch	(a program written by Jason Woofenden)\n%d %d\n255\n", width, height);

	while(pixbuf < pixend)
	putchar((char)*pixbuf++);
}
