#include <math.h>
#include <stdio.h>
#include <SDL.h>

#include "koch.h"
#include "sdl.h"

#include "backend.h"
#define __KOCH_GUI_C_
#include "backend.c"

int center_x, center_y;
double angle = 0;
double radius = 10;

uint32_t red;
uint32_t line_color;

/* How close you have to click to get the point */
#define POINTSIZE 5

/* Max number of points allowed */
typedef struct
{
  int x, y;
} point;

point points[MAXPOINTS];
int curPoint = 0;
point *selected = 0;

/* BUFSIZE is how many floats. */
#define MAXRENDER1 4000
#define MAXRENDER2 80000
#define BUFSIZE ((MAXRENDER2+MAXPOINTS) * 2 + 2)

/* Called before the program quits */
void quit_now()
{
  int c;
  if(curPoint)
    {
      for(c = 0; c < curPoint; ++c)
	{
	  printf(" %d %d ", points[c].x, points[c].y);
	}
      putchar('\n');
    }
}

void render_fbuf(float* buf, int count)
{
  int iterations, maxIterations, bufSize;
  float x, y, theta, r, newTheta, newR;
  int x1, y1, x2, y2;
  
  x = points[curPoint-1].x - points[0].x;
  y = points[curPoint-1].y - points[0].y;
  r = sqrt(x*x + y*y);
  theta = patan(x, y);
  x = points[0].x;
  y = points[0].y;

  safe_move_to(x, y);

  x2 = x;
  y2 = y;

  while(count--)
    {
      x1 = x2;
      y1 = y2;

      newR = *buf++ * r;
      newTheta = *buf++ + theta;

      x += cos(newTheta) * newR;
      y += sin(newTheta) * newR;

      x2 = x;
      y2 = y;

      safe_line_to(x, y, line_color);
    }
}


void draw_koch(int count)
{
  float x,y;
  int c;
  float *bufCur, *newBufCur;
  static float* buf = 0;

  if(buf == 0)
    {
      buf = (float*)malloc(8 + BUFSIZE * sizeof(float));
      if(!buf)
	{
	  fprintf(stderr, "Could not allocate %d bytes\n", BUFSIZE * sizeof(float));
	  quit_now();
	  exit(1);
	}
    }


  /* copy points to buf */
  bufCur = buf;
  for(c = 0; c < curPoint; ++c)
    {
      *bufCur++ = (float)points[c].x;
      *bufCur++ = (float)points[c].y;
    }

  /* convert points to relative, noramalized format for koch */
  rec_to_norm_polar(buf, curPoint);


  x = points[curPoint-1].x - points[0].x;
  y = points[curPoint-1].y - points[0].y; 
  /* recursively replace */
  c = koch(buf, curPoint-1, count, 2.1 / sqrt(x*x + y*y));

  /* draw it on the screen (and in the offscreen pixmap) */
  render_fbuf(buf, c);

}  


void draw_points()
{
	int c;

	for(c = 0; c < curPoint; ++c)
	{
		safe_move_to(points[c].x - POINTSIZE, points[c].y - POINTSIZE);
		safe_line_to(points[c].x + POINTSIZE, points[c].y - POINTSIZE, red);
		safe_line_to(points[c].x + POINTSIZE, points[c].y + POINTSIZE, red);
		safe_line_to(points[c].x - POINTSIZE, points[c].y + POINTSIZE, red);
		safe_line_to(points[c].x - POINTSIZE, points[c].y - POINTSIZE, red);
	}

	for(c = 1; c < curPoint; ++c)
	{
		safe_move_to(points[c-1].x, points[c-1].y);
		safe_line_to(points[c].x, points[c].y, red);
	}

	update_screen();
}

void redraw(int points)
{
	int c;

	clear_buffer();

	if(curPoint > 2)
		draw_koch(points);

	draw_points();
}

/* return point near (x,y) or nil */
point* find_point(int x, int y)
{
  int c;
  int dx, dy;
  for(c = 0; c < curPoint; ++c)
    {
      dx = points[c].x - x; 
      dy = points[c].y - y;

      if(sqrt((float)(dx*dx + dy*dy)) <= POINTSIZE)
	return &points[c];
    }
  return 0;
}

/* Called when the user clicks on the drawing area. */
/* add apoint where they clicked */
/* return true if the display should be updated */
int click(int x, int y)
{
  selected = find_point(x, y);
  if(!selected)
    {
      if(curPoint+1 < MAXPOINTS)
	{
	  points[curPoint].x = x;
	  points[curPoint].y = y;
	  selected = &points[curPoint];
	  ++curPoint;
	  return 1;
	}
    }

  return 0;
}

/* Called when the user right-clicks on the drawing area. */
/* remove the point they clicked od */
/* return true if the display should be updated */
int right_click(int x, int y)
{
	selected = find_point(x, y);
	if(selected)
	{
		int right;

		/* shift points past this one down */
		for(right = (selected - points); right < curPoint - 1; ++right)
		{
			points[right].x = points[right + 1].x;
			points[right].y = points[right + 1].y;
		}
		--curPoint;

		selected = 0;

		return 1;
	}

return 0;
}

/* Called when the user drags in the drawing area. */
/* return true if the display should be updated */
int drag(int x, int y)
{
  if(selected)
    {
      selected->x = x;
      selected->y = y;
      return 1;
    }

  return 0;
}

void usage(char *argv[])
{
  fprintf(stderr, "usage: %s [x y [x y [...]]]\n", argv[0]);
  exit(1);
}

void event_loop()
{
	render_args rargs;
	SDL_Event event;
	int dragged, dragx, dragy;
	int resize = 0, resizew = 640, resizeh = 480;
	int screen_changed = 1;
	int points_visible = 1, moved = 0;
	int idle = 0;
	int pixbuf_size = 0;
	int c;
	float *pixbuf = 0;
	float *buf = 0;
	float *original = (float*)malloc(MAXPOINTS*2 * sizeof(float)); /* fix, I only need this because backend uses a different format for the x,y pairs in the points for the original. */
	if(!original)
	{
		exit(12);
	}

	for(;;)
	{

		for(;;)
		{
			//if(idle || dragged || resize || screen_changed)
			{
				if(SDL_PollEvent(&event) == 0) break;
			}
			/* FIXME I have commented this out because it does not stop waiting when the mouse moves (and isn't clicking
			else // if we have nothing to do then wait
			{
				if(SDL_WaitEvent(&event) == 0)
				{
					// This should never happen
					printf("SDL_WaitEvent error: %s\n", SDL_GetError());
					return;
				}
			} */

			switch (event.type)
			{
				case SDL_VIDEORESIZE:
					resize = 1;
					resizeh = event.resize.h;
					resizew = event.resize.w;
					fprintf(stderr, "Resize event...  ");
					break;
				case SDL_MOUSEBUTTONDOWN:
					if(!resize)
						if(event.button.button == SDL_BUTTON_LEFT)
							screen_changed |= click(event.button.x, event.button.y);
						else if(event.button.button == SDL_BUTTON_RIGHT)
							screen_changed |= right_click(event.button.x, event.button.y);
					break;

				case SDL_MOUSEMOTION:
					if(event.motion.state & SDL_BUTTON(SDL_BUTTON_LEFT))
					{
						dragged = 1;
						dragx = event.motion.x;
						dragy = event.motion.y;
					}
					moved = 1;
					break;

				case SDL_QUIT:
					/*fprintf(stderr, "Quitting on request\n");*/
					return;
			}
		}
		
		if(resize)
		{
			resize = 0;
			gfx_resize(resizew, resizeh);
			screen_changed = 1;
			fprintf(stderr, "Resized\n");
		}

		if(dragged)
		{
			dragged = 0;
			screen_changed |= drag(dragx, dragy);
		}

		if(screen_changed)
		{
			redraw(MAXRENDER1);
			points_visible = 1;
			screen_changed = 0;
			moved = 0;
			idle = 1;
		}
		else
		{
			if(moved && !points_visible)
			{
				draw_points();
			}
			
			if(idle)
			{
				if(curPoint < 3)
				{
					idle = 0;
				}
				else
				{
					switch (idle)
					{
						/* draw with more lines */
						case 1:
							redraw(MAXRENDER2);
							++idle;
							break;

						/* allocate buffers */
						case 2:
							++idle;

							if(buf == 0)
							{
								buf = (float*) malloc(4 * 10000 * sizeof(float));
								if(buf == 0)
								{
									fprintf(stderr, "Couldn't allocate %d bytes for render stack.\n", 4 * 10000 * sizeof(float));
									exit(5);
								}

								break;
							}
							/* if it's already allocated then fall through to the next step. */

						case 3:
							++idle;

							if(pixbuf_size < resizeh * resizew)
							{
								if(pixbuf)
									free(pixbuf);
								
								pixbuf_size = resizeh * resizew;
								pixbuf = (float*) malloc(pixbuf_size * sizeof(float));
								if(pixbuf == 0)
								{
									fprintf(stderr, "Couldn't allocate %d bytes for the renderer's screenbuffer.\n", 4 * 10000 * sizeof(float));
									exit(5);
								}
								
								break;
							}
							/* if it's already allocated (and big enough) then fall through to the next step. */

						/* setup renderer */
						case 4:
							++idle;

							for(c = 0; c < MAXPOINTS; ++c)
							{
								original[c*2] = points[c].x;
								original[c*2 + 1] = points[c].y;
							}

							rargs.g_width = resizew;
							rargs.g_height = resizeh;
							rargs.g_max_length = RESOLUTION;
							rargs.g_pixbuf = pixbuf;
							rargs.g_original_size = curPoint - 1;
							rargs.tos = rargs.buf = buf;
							rargs.g_original = original;
							rargs.count = 0;

							if(rargs_setup(&rargs))
							{
								idle = 0;
							};

							break;
							
						/* render */
						case 5:
							if(render_3(&rargs, 10000))
								++idle;
							break;

						/* adjust brightness */
						case 6:
							adjust_brightness(rargs.g_pixbuf, rargs.g_width, rargs.g_height);
							++idle;
							break;
						
						/* blit */
						case 7:
							blit_floats(rargs.g_pixbuf, rargs.g_width, rargs.g_height);
							update_screen();
							points_visible = 0;
							moved = 0;
							idle = 0;
							break;
					}
				}
			}
		}
	}
}
	
	

int main(int argc, char *argv[])
{
	int c;
	if(argc > MAXPOINTS*2 + 1 || !(argc % 2))
	usage(argv);

	if(argc > 1)
	{
		for(curPoint = 0, c = 1; curPoint < MAXPOINTS && c < argc; ++curPoint)
		{
			points[curPoint].x = atoi(argv[c]);
			++c;
			points[curPoint].y = atoi(argv[c]);
			++c;
		}
	}

	gfx_init();
	atexit(SDL_Quit);

	red = rgb_to_pixel(255,0,0);
	line_color = rgb_to_pixel(155,155,155);
	
	event_loop();

	quit_now();

	return 0;
}

