#include <math.h>

#include "koch.h"

/* koch():  recurse: relative polar -> relative polar
 *
 * INSTRUCTIONS:
 * 1) create a buffer to hold at least doneSize*2 floats.
 * 2) put the original line segments at the begining of the buffer as 
 *    a relative polar, normalized (see rec_to_norm_polar()).
 *    That is, an array of origSize [r,theta] pairs
 *    (where theta is to be added the the direction (in radians) of
 *    this segment, and r is the length of the segment.)
 * 3) call this function, passing the following:
 *
 * PARAMETERS & RETURN
 *    koch: the buffer. THIS IS MODIFIED. It must be at least
 *       doneSize*2 floats long. The result of the recursion will go
 *       in this buffer. The number of pairs written is returned.
 *    originalSize: Number of segments you put in koch.
 *    doneSize: Max number of segments this function will calculate.
 *       Must be one more than an even multiple of (originalSize-1).
 *    smallEnough: segment length at which this function considers it
 *       small enough, and won't recurse in any further.
 *    RETURN: number of segments written to buffer (koch)
 */
int koch(float *koch, int originalSize, int doneSize, float smallEnough)
{
  float *curNew, *curOrig, *end = koch + doneSize*2;
  float *lastEnd; /* end of new last time */
  float original[MAXPOINTS*2];
  float *curOriginal, *originalMax = original + originalSize*2; /* sorry I wasn't too creative with names. don't get this confused with curOrig */
  float x;

  /* make copy of original */
  memcpy(original, koch, originalSize*2 * sizeof(float));

  /* loop setup */
  curNew = koch + originalSize*2; /* so memcpy will work right */
  lastEnd = koch; curOrig = end; /* so loop will start */

  /* foreach(iteration) */
  while(curNew > lastEnd && curNew < curOrig)
    {
      lastEnd = curNew;

      /* copy segments to end of buffer, and recalculate buffer positions */
      curOrig = end - (curNew - koch);
      memmove(curOrig, koch, (curNew - koch) * sizeof(float));
      curNew = koch;

      /* foreach(segment from last time) */
      while(curOrig < end && curNew < curOrig)
	{
	  if(curOrig[0] > smallEnough)
	    {
	      /* foreach(segment in original) */
	      for(curOriginal = original; curOriginal < originalMax;)
		{
		  /* original is normalized, so we just multiply the length */
		  *curNew++ = curOrig[0] * *curOriginal++;

		  /* add theta to that of last iteration */
		  *curNew++ = curOrig[1] + *curOriginal++;
		}
	      curOrig += 2;
	    }
	  else
	    {
	      /* if this segment is already long enough then just copy in it */
	      *curNew++ = *curOrig++;
	      *curNew++ = *curOrig++;
	    }
	}

    }

  if(curNew >= curOrig)
    return doneSize;

  /* return number of segments */
  return (curNew - koch)/2; 
}




/* This changes koch so it is a relative, normalized polar array as is needed to
 * call koch(). See the comments above koch() for an overview
 *
 *    PARAMETERS
 * koch: an array of size [x,y] pairs
 * size: the number of [x,y] pairs
 */
void rec_to_norm_polar(float *koch, int size)
{
  float *curKoch, *end;
  float x, y, r, theta;
  int c;

  /* shift original to origin */
  x = koch[0]; y = koch[1];
  curKoch = koch;
  end = koch + size*2;
  while(curKoch < end)
    {
      *curKoch++ -= x;
      *curKoch++ -= y;
    }

  /* convert to polar with r=1 */
  x = koch[(size-1)*2];
  y = koch[(size-1)*2 + 1];
  r = sqrt(x*x + y*y);
  theta = patan(x,y);
  x = y = 0;
  for(c = 0; c < size*2 - 2; ++c)
    {
      x = koch[c + 2] - koch[c];
      y = koch[c + 3] - koch[c+1];
      
      koch[c] = sqrt(x*x + y*y)/r;
      ++c;
      koch[c] = patan(x, y) - theta;
    }
}


/* sign-preserving tan */
float patan(float x, float y)
{
	if(x == 0.0)
	{
		if(y > 0)
			return M_PI/2.0;
		else
			return M_PI*1.5;
	}
  /* if(x < .000000001)
    return y > 0 ? M_PI / 2.0 : M_PI / -2.0; */
  return ((x > 0) ? atan(y/x) : M_PI+atan(y/x));
}

