/* sdl.c - functions to interface with SDL library.
   Copyright (C) 2002 Joshua I. Grams

   This file is part of meissa.

   meissa is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   meissa is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with meissa; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA. */

#include <SDL.h>
#include <SDL_endian.h> /* Used for the endian-dependent 24 bpp mode */
#include "sdl.h"

struct turtle {
	/* current position */
	uint16_t x, y;
	uint8_t *cur;
	uint8_t inside;
};

struct line {
	int16_t dmaj, dmin, half_dmaj;
	int32_t mv_maj, mv_min;
	int t;
	uint8_t *cur;
	uint8_t *end;
};

uint32_t greys[256];
SDL_Color colors[256];

SDL_Surface *screen;
struct turtle scr;

/* set line_to function according to pixel size
 * open sdl window */
int gfx_init(void) /* returns -1 on failure, 0 on success */
{
	int i;

	/* Fill palette with greys */
	for(i = 0; i < 256; ++i)
	{
		colors[i].r=i;
		colors[i].g=i;
		colors[i].b=i;
	}
	/* put red in the palette */
	colors[254].r = 255;
	colors[254].g = 0;
	colors[254].b = 0;

	if(SDL_Init(SDL_INIT_VIDEO) < 0)
	{
		fprintf(stderr, "Could not initialize SDL: %s", (int)SDL_GetError());
	}
	
	//gfx_resize(1152, 850);
	gfx_resize(640, 480);
	
	/* It's all a big lie!!!!! this one is red. */
	greys[255] = SDL_MapRGB(screen->format, 255, 0, 0);

	switch(screen->format->BytesPerPixel)
	{
		case 1:
			line_to = line_to_8;
			break;
		case 2:
			line_to = line_to_16;
			break;
		case 3:
			if(SDL_BYTEORDER == SDL_BIG_ENDIAN)
				line_to = line_to_b24;
			else
				line_to = line_to_s24;
			break;
		case 4:
			line_to = line_to_32;
			break;
	}

	/* Lock the screen for direct access to the pixels */
	if ( SDL_MUSTLOCK(screen) ) {
		if ( SDL_LockSurface(screen) < 0 ) {
			fprintf(stderr, "Can't lock screen: %s\n", SDL_GetError());
			return;
		}
	}

	/* allocate 8 bit screen buffer for greyscale art */
	/*screen = SDL_CreateRGBSurface(SDL_HWSURFACE, screen->w, screen->h, 8, 0xff, 0xff, 0xff, 0);
	if(screen == NULL)
	{
		fprintf(stderr, "CreateRGBSurface failed: %s\n", SDL_GetError());
		exit(1);
	}*/

}

void gfx_resize(int w, int h)
{
	int i;

	screen = SDL_SetVideoMode(w, h, 0, SDL_RESIZABLE | SDL_HWPALETTE);
	if(screen == 0)
	{
		fprintf(stderr, "Couldn't get video mode: %s", (int)SDL_GetError());
	}

	if(screen->format->BytesPerPixel == 1)
	{
		i = SDL_SetPalette(screen, SDL_LOGPAL|SDL_PHYSPAL, colors, 0, 256);
		if(i == 0)
			fprintf(stderr, "Warning: couldn't set whole color palette for screen\n");
	}

	for(i = 0; i < 255; ++i)
	{
		greys[i] = SDL_MapRGB(screen->format, i, i, i);
	}
}

int gfx_exit(int ignored) /* ditto */
{
	SDL_Quit();
}

void update_screen()
{
	if ( SDL_MUSTLOCK(screen) ) {
		SDL_UnlockSurface(screen);
	}
	/* Update just the part of the display that we've changed */

	/*SDL_BlitSurface(offscreen, 0, screen, 0);*/
	SDL_UpdateRect(screen, 0, 0, screen->w, screen->h);

	/* Lock the screen for direct access to the pixels */
	if ( SDL_MUSTLOCK(screen) ) {
		if ( SDL_LockSurface(screen) < 0 ) {
			fprintf(stderr, "Can't lock screen: %s\n", SDL_GetError());
			return;
		}
	}

	/*clear_buffer();*/
}

void clear_buffer()
{
	memset(screen->pixels, 0, screen->pitch * screen->h);
}

uint32_t rgb_to_pixel(uint8_t r, uint8_t g, uint8_t b)
{
	return SDL_MapRGB(screen->format, r, g, b);
}

void move_to(uint16_t x, uint16_t y)
{
	scr.cur = (uint8_t*)(screen->pixels) + (y * screen->pitch) + x * screen->format->BytesPerPixel;
	scr.x = x;
	scr.y = y;
}

void safe_move_to(uint16_t x, uint16_t y)
{
	scr.inside = (y < (uint16_t)(screen->h) && x < (uint16_t)(screen->w));
	move_to(x,y);
}

void setup_line(uint16_t x, uint16_t y, uint32_t color, struct line *l)
{
	int16_t dt;
	int32_t mv_t;

	/* get difference and increment for x */
	l->dmaj = x - scr.x;
	if(l->dmaj < 0) {
		l->mv_maj = -screen->format->BytesPerPixel;
		l->dmaj = -l->dmaj;
	} else {
		l->mv_maj = screen->format->BytesPerPixel;
	}
	/* get difference and increment for y */
	l->dmin = y - scr.y;
	if(l->dmin < 0) {
		l->mv_min = -screen->pitch;
		l->dmin = -l->dmin;
	} else {
		l->mv_min = screen->pitch;
	}
	/* if y difference greater than x difference, swap them */
	if(l->dmin > l->dmaj) {
		dt = l->dmaj; mv_t = l->mv_maj;
		l->dmaj = l->dmin; l->mv_maj = l->mv_min;
		l->dmin = dt; l->mv_min = mv_t;
	}
	/* finish setting up values */
	l->half_dmaj = l->dmaj;
	l->dmaj = 2 * l->dmaj;
	l->dmin = 2 * l->dmin;
	l->t = 0;
	l->cur = scr.cur;
	l->end = (uint8_t*)(screen->pixels) + (y * screen->pitch) + x * screen->format->BytesPerPixel;
	scr.x = x;
	scr.y = y;
	scr.cur = l->end;
}

void (*line_to)(uint16_t x, uint16_t y, uint32_t color);

#define LINETOFUNC(func_name) \
	void func_name (uint16_t x, uint16_t y, uint32_t color) \
	{ \
		struct line l; \
		setup_line(x, y, color, &l); \
		/* loop and draw pixels */ \
		do { \
			SETPIXEL; \
			l.t += l.dmin; \
			if(l.t > l.half_dmaj) { l.t -= l.dmaj; l.cur += l.mv_min; } \
			l.cur += l.mv_maj; \
		} while(l.cur != l.end); \
		SETPIXEL; \
	}

#define SETPIXEL *((uint8_t*)(l.cur)) = color
LINETOFUNC(line_to_8);

#undef SETPIXEL
#define SETPIXEL *((uint16_t*)(l.cur)) = color
LINETOFUNC(line_to_16);

#undef SETPIXEL
#define SETPIXEL \
	{ \
		l.cur[2] = color; \
		l.cur[1] = color >> 8; \
		l.cur[0] = color >> 16; \
	}
LINETOFUNC(line_to_b24);

#undef SETPIXEL
#define SETPIXEL \
	{ \
		l.cur[0] = color; \
		l.cur[1] = color >> 8; \
		l.cur[2] = color >> 16; \
	}
LINETOFUNC(line_to_s24);

#undef SETPIXEL
#define SETPIXEL *((uint32_t*)(l.cur)) = color
LINETOFUNC(line_to_32);


void safe_line_to(uint16_t x, uint16_t y, uint32_t color)
{
	if(y >= (uint16_t)(screen->h) || x >= (uint16_t)(screen->w))
	{
		scr.x = x;
		scr.y = y;
		scr.inside = 0;
		return;
	}

	if(!scr.inside)
	{
		move_to(x,y);
		scr.inside = 0xFF;
		return;
	}

	if(x != scr.x || y != scr.y)
		line_to(x, y, color);
}

void blit_floats(float* pix, int width, int height)
{
	int x, y;
	uint8_t *spix;
	uint32_t pixel;

	for(y = 0; y < height; ++y)
	{
		spix = (uint8_t*)screen->pixels + (y * screen->pitch);
		for(x = 0; x < width; ++x)
		{
			pixel = greys[(int)*pix++];
			switch (screen->format->BytesPerPixel)
			{
				case 1:
					*((uint8_t*)spix) = pixel;
					break;
				case 2:
					*((uint16_t*)spix) = pixel;
					break;
				case 3:
					if(SDL_BYTEORDER == SDL_BIG_ENDIAN)
					{
						spix[2] = pixel;
						spix[1] = pixel >> 8;
						spix[0] = pixel >> 16;
					}
					else
					{
						spix[0] = pixel;
						spix[1] = pixel >> 8;
						spix[2] = pixel >> 16;
					}
				case 4:
					*((uint32_t*)spix) = pixel;
					break;
				default:
					exit(48);
			}

			spix += screen->format->BytesPerPixel;
		}
	}
}
